# -*- coding: utf-8 -*-

import argparse, ctypes, html, json, os, platform, re, subprocess, sys, tempfile, textwrap, time
from functools import lru_cache
from os.path import exists

class params:
    pass

def main(*,
         aws_route53_zone_id,
         code,
         config,
         config_script_filename,
         issues_uri,
         launched_from,
         livenessuri,
         longname,
         oneline_setup,
         port,
         repo_dir_for_config_script,
         repo_uri,
         subnet,
         zone,
         ):
    # Validate received params
    assert isinstance(aws_route53_zone_id, str)
    assert re.match(r"^[a-z]{1,10}$", code)
    assert callable(config)
    assert re.match(r"^[a-z]+\.py$", config_script_filename)
    assert re.match(r"^https://[a-z]", issues_uri)
    assert launched_from in ["deployed", "repo"]
    assert re.match(r"^http://", livenessuri)
    assert re.match(r"^[A-Za-z0-9 ]{1,50}$", longname)
    assert isinstance(oneline_setup, str) and config_script_filename in oneline_setup
    assert isinstance(port, int) and 1024 <= port <= 65535
    assert re.match(r"^([a-z0-9]+/)*$", repo_dir_for_config_script)
    assert re.match(r"^https://[a-z]", repo_uri)
    assert isinstance(subnet, int) and 0 <= subnet <= 255
    if subnet <= 10: raise ValueError("A subnet in the 0-10 range is inadvisable, as it might clash with common configurations.")
    assert re.match(dns_name_pattern, zone)
    # Store received params
    params.aws_route53_zone_id=aws_route53_zone_id
    params.code=code
    params.config=config
    params.config_script_filename=config_script_filename
    params.issues_uri=issues_uri
    params.launched_from=launched_from
    params.livenessuri=livenessuri
    params.longname=longname
    params.oneline_setup=oneline_setup
    params.port=port
    params.repo_dir_for_config_script=repo_dir_for_config_script
    params.repo_uri=repo_uri
    params.subnet=subnet
    params.zone=zone
    # Calculated params
    wgdir="/etc/wireguard-tmp"
    interface=f"wgt-{code}"
    params.config_file=f"{wgdir}/{interface}.conf"
    params.interface=interface
    params.privkeyfile=f"{wgdir}/keys/{interface}.privatekey"
    params.pubkeyfile=f"{wgdir}/keys/{interface}.publickey"
    params.setupstatefile=f"{wgdir}/keys/{interface}.json"

    # Parse arguments
    parser = argparse.ArgumentParser(description=f"One-line setup for {params.longname}: {params.oneline_setup}")
    parser.add_argument(
        '--setup',
        action='store_true',
        help='Setup or refresh the VPN connection on this machine.')
    parser.add_argument(
        '--setup-other',
        action='store_true',
        help='Get interactive help creating the config line for another host.')
    parser.add_argument(
        '--test',
        action='store_true',
        help='Test the VPN configuration for internal consistency.')
    parser.add_argument(
        '--config',
        metavar='SPEC',
        type=str,
        help=f'Generate configuration for a host on {params.longname} and print to stdout. The host is specified by public_key, IP address on VPN, DNS label/domain on VPN, public domain or description.')
    parser.add_argument(
        '--config-tarball',
        metavar='SPEC',
        type=str,
        help=f'Produce a tarball with a README and configuration file for a host on {params.longname}.')
    parser.add_argument(
        '--update-dns',
        action='store_true',
        help='Use aws CLI tool to update DNS zones.')
    parser.add_argument(
        '--render-config',
        metavar='FORMAT',
        choices=["html", "json"],
        help=f'Render a list of hosts on {params.longname} in the specified format (html or json).')
    args = parser.parse_args()

    do_default_action = (not args.setup and
                         not args.setup_other and
                         not args.test and
                         not args.config and
                         not args.config_tarball and
                         not args.update_dns and
                         not args.render_config)
    do_setup = args.setup or (do_default_action and launched_from == "deployed")
    do_help = do_default_action and launched_from == "repo" # argparse will handle explicit --help

    if do_help:
        parser.print_help()
        sys.exit(0)

    # We always load and check the configuration, no matter the arguments.
    try:
        load_and_check_config()
    except ValueError as e:
        print(f"There's a problem with your configuration: {e.args[0]}")
        suggest("- This should not happen when running the one-line setup. If it does, please",
                f"  file a bug report at {params.issues_uri}")
        suggest(f"- If this happens after you've made local edits to {params.config_script_filename}, you need to fix",
                "  the problem before submitting a PR.")
        exit_with_suggestions(1)

    if len([x for x in [args.setup,
                        args.setup_other,
                        args.test,
                        args.config,
                        args.update_dns,
                        args.render_config] if x])>1:
        print("You can only use one command flag.")
        suggest(f"- Run {params.config_script_filename} --help")
        exit_with_suggestions(1)

    if do_setup:
        setup()
        exit_with_suggestions(0)

    if args.setup_other:
        if not (os.isatty(0) and os.isatty(1)):
            print("Cannot use terminal for interactive questions.")
            if params.launched_from == "deployed":
                suggest("- Try checking out the puppet repo in order to run --setup-other.")
            else:
                suggest("- Try running this script in an interactive shell.")
            exit_with_suggestions(1)
        interactive_setup_help_pr(public_key=None, on_this_machine=False)
        exit_with_suggestions(0)

    if args.test:
        print(f"VPN configuration for {params.longname} is internally consistent.")
        exit_with_suggestions(0)

    if args.config:
        spec=args.config
        host=get_host(spec)
        if host == None or not host.enabled:
            print(f"No configuration found for this SPEC: {spec}")
            exit_with_suggestions(1)
        config_text=wireguard_config(host.public_key, '<PRIVATE_KEY>')
        print(config_text)
        sys.exit(0)

    if args.config_tarball:
        spec=args.config_tarball
        host=get_host(spec)
        if host == None or not host.enabled:
            print(f"No configuration found for this SPEC: {spec}")
            exit_with_suggestions(1)
        make_config_tarball(host)
        sys.exit(0)

    if args.update_dns:
        try:
            update_dns()
        except subprocess.CalledProcessError as e:
            if e.returncode in [253, 254, 255]:
                print(f"For automation purposes, you can store AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY in {params.setupstatefile}")
                exit_with_suggestions(1)
            raise e
        sys.exit(0)

    if args.render_config:
        render_config(render_format=args.render_config)
        sys.exit(0)

    print("Bug")
    sys.exit(1)

## Functions for loading the config, checking it for internal consistency, and fetching a config item

segments = []
current_segment = None
host_configs = []
host_config_map = {}

# A valid and canonical base64-encoded byte sequence can be checked by this
# regex:
# ^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/][AQgw]==|[A-Za-z0-9+/]{2}[AEIMQUYcgkosw048]=)?$
# Since wireguard keys (both public and private) are 256 bit values, and 256 =
# 6*43-2, they should also match
# ^[^=]{43}=$
# The intersection of those two regexps is:
# ^[A-Za-z0-9+/]{42}[AEIMQUYcgkosw048]=$
wg_key_pattern = re.compile(r'^[A-Za-z0-9+/]{42}[AEIMQUYcgkosw048]=$')

dns_label_pattern = re.compile(r'^(?!-)[a-z0-9-]{1,63}(?<!-)$')
dns_name_pattern = re.compile(r'^(?=.{1,253}$)((?!-)[a-z0-9-]{1,63}(?<!-)\.)+[a-z]{2,63}$')
description_pattern = re.compile(r'^(?! )[A-Za-zÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ 0-9,._=&/*():#;-]+(?<! )$')
person_pattern = re.compile(r'^(?! )[A-Za-zÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ ]+(?<! )$')
device_pattern = re.compile(r'^[a-z0-9-]+$')

def load_and_check_config(extra=None):
    # Reset the data structures since this could be called more than once.
    global segments, host_configs, host_config_map
    segments = []
    host_configs = []
    host_config_map = {}

    # Load configuration into segments and host_configs and check consistency of
    # individual records.
    params.config()
    if extra: extra()

    # Check foreign constraints

    assert_unique([segment.X for segment in segments],
                  "Duplicate segment X value detected: %s")
    available_segments = [segment.X for segment in segments if segment.available]
    for host in host_configs:
        if not host.enabled: continue
        if host.X not in available_segments:
            raise ValueError(f"Segment X={host.X} is not available.")
    assert_unique([(host.X, host.Y) for host in host_configs
                   if host.X!=None and host.Y!=None],
                  "Duplicate host configuration for X=%d and Y=%d")
    assert_unique([host.public_key for host in host_configs
                   if host.public_key!=None],
                  "Duplicate public key detected: %s")
    assert_unique([host.dns_on_vpn for host in host_configs
                   if host.dns_on_vpn!=None],
                  "Duplicate dns_on_vpn name detected: %s")
    assert_unique([host.public_dns for host in host_configs
                   if host.public_dns!=None],
                  "Duplicate public DNS name detected: %s")
    disabled_people = [host.person for host in host_configs
                       if host.person!=None and host.device==None and not host.enabled]
    for host in host_configs:
        if host.enabled:
            if host.person in disabled_people:
                raise ValueError(f"Host configured for person {host.person} is not allowed.")
    assert_unique([(host.person,host.device) for host in host_configs
                   if host.person!=None and host.device!=None],
                  "Duplicate person/device combination: %s, %s")
    assert_unique([host.description for host in host_configs
                   if host.description!=None],
                  "Duplicate description detected: %s")

    # Build host_config_map and check for duplicate SPECs

    # We could have done this in host(), but we want to do it after checking
    # foreign constraints in order to give the user more friendly error
    # messages.

    for host in host_configs:
        # As keys, use:
        # - public_key
        # - VPN IP address
        # - VPN DNS label
        # - VPN DNS domains
        # - public DNS
        # - description
        keys = [
            host.public_key,
            ip(host) if host.X is not None else None,
            host.dns_on_vpn,
            f"{host.dns_on_vpn}.{params.zone}" if host.dns_on_vpn else None,
            host.public_dns,
            host.description]
        # Deduplicate and remove None
        keys = list(dict.fromkeys([key for key in keys if key!=None]))
        # Check that no other host has used any of these keys
        for key in keys:
            if key in host_config_map:
                print(f"Duplicate SPEC: {key}")
                print("As SPEC we use public_key, IP address on VPN, DNS label/domain on VPN,")
                print("public_dns and description.")
                print("First host:")
                print(host_config_map[key])
                print("Second host:")
                print(host)
                raise ValueError(f"Duplicate SPEC: {key}")
            host_config_map[key] = host

class segment:
    def __init__(self, *, X, available, description, personal=False):
        """
        Used in the configuration to define a network segment.
        Arguments:
        - X (required): The network segment, 3rd part of IP address.
        - available (required): True if hosts can be assigned in this segment.
          Segments not mentioned will be unavailable, but one might want to
          mention a segment that is reserved but not currently used, in which
          case available=False is useful.
        - description (required): A string describing the intended use of this
          segment.
        - personal (optional): If True, only allow personal devices in this
          segment. Default is False, to not allow any personal devices.
        """
        if not isinstance(X, int) or not 0 <= X <= 255:
            raise ValueError("X must be in the range 0-255.")
        if not isinstance(available, bool):
            raise ValueError("available must be a boolean.")
        if not isinstance(description, str) or not description_pattern.match(description):
            raise ValueError("description must be a non-empty string with sane punctuation.")
        if not isinstance(personal, bool):
            raise ValueError("personal must be a boolean.")
        self.X = X
        self.available = available
        self.description = description
        self.personal = personal
        segments.append(self)
    def __enter__(self):
        global current_segment
        current_segment = self.X
    def __exit__(self, exc_type, exc_value, traceback):
        global current_segment
        current_segment = None
        return False

class host_definition:
    def __init__(self, *, X, Y, public_key, dns_on_vpn, public_dns, description, person, device, prefix_length, enabled):
        """Used in public_server, private_server, personal, and disable functions."""
        global current_segment
        if not enabled and X==None and Y==None: pass
        else:
            if not isinstance(X, int) or not 0 <= X <= 255:
                raise ValueError(f"X={X} must be in the range 0-255.")
            if not isinstance(Y, int) or not 0 <= Y <= 254:
                raise ValueError(f"Y={Y} must be in the range 0-254.")
            if not X==current_segment:
                raise ValueError(f"Host with segment X={X} must be defined under the same segment.")
        if not enabled and public_key==None: pass
        else:
            if not isinstance(public_key, str) or not wg_key_pattern.match(public_key):
                raise ValueError(f"public_key={public_key} is not a valid WireGuard public key.")
        if not dns_on_vpn==None:
            if not isinstance(dns_on_vpn, str) or not dns_label_pattern.match(dns_on_vpn):
                raise ValueError("dns_on_vpn must be None or a string matching the DNS label format.")
        if not public_dns==None:
            if not isinstance(public_dns, str) or not dns_name_pattern.match(public_dns):
                raise ValueError("public_dns must be None or a string matching the DNS name format.")
        if not enabled and description==None: pass
        else:
            if not isinstance(description, str) or not description_pattern.match(description):
                raise ValueError("description must be a non-empty string with sane punctuation.")
        if not person==None:
            if not isinstance(person, str) or not person_pattern.match(person):
                raise ValueError("person must be a non-empty string containing only Latin-1 letters and spaces.")
        if not device==None:
            if not isinstance(device, str) or not device_pattern.match(device):
                raise ValueError("device must be a non-empty string containing only a-z, 0-9 and spaces.")
        if person==None and not device==None:
            raise ValueError(f"Person must be given if device is given: {device}")
        if not person==None and device==None:
            raise ValueError(f"Device must be given if person is given: {person}")
        if not prefix_length==None and (not isinstance(prefix_length, int) or not 16 <= prefix_length <= 31):
            raise ValueError("prefix_length must be an integer in the range 16-31.")
        if not isinstance(enabled, bool):
            raise ValueError("enabled must be a boolean.")
        self.X = X
        self.Y = Y
        self.public_key = public_key
        self.dns_on_vpn = dns_on_vpn
        self.public_dns = public_dns
        self.description = description
        self.person = person
        self.device = device
        self.prefix_length = prefix_length
        self.enabled = enabled
        global host_configs
        host_configs.append(self)
    def __repr__(self):
        return "\n".join([
            f"- {key}={repr(val)}"
            for key, val in [("X", self.X),
                             ("Y", self.Y),
                             ("public_key", self.public_key),
                             ("dns_on_vpn", self.dns_on_vpn),
                             ("public_dns", self.public_dns),
                             ("description", self.description),
                             ("person", self.person),
                             ("device", self.device),
                             ("prefix_length", self.prefix_length),
                             ("enabled", self.enabled)]])

def ip(host):
    return f"10.{params.subnet}.{host.X}.{host.Y}"

"""
Functions to define a host:
- public_server: Servers that are accessible from the internet
- private_server: Servers that are not accessible from the internet
- personal: Personal devices

Arguments are all named (no positional):
- X (required): The network segment, 3rd part of IP address. Choose a segment
  that matches the purpose stated in the segment definition.
- Y (required): 4th part of IP address. Choose the next available Y in the range
  2-254 unless there's a good reason to use 0 or 1 (which traditionally are
  reserved for network equipment).
- public_key (required): WireGuard public key.
- dns_on_vpn (optional): DNS label to use on the VPN. This creates a DNS record
  <LABEL>.<DNS-ZONE> for the host.
- public_dns (required for public_server, otherwise not given): The domain
  name for resolving the host on the internet.
- prefix_length (optional for public_server, otherwise not given): The
  IP prefix length for the subnet a wireguard bounce server should route
  for. If not given, the host will not act as a bounce server.
- description (not given for personal, optional if dns_on_vpn is given,
  otherwise required): A string describing the host.
- person (required for personal, otherwise not given): The name of the
  device owner.
- device (optional for personal, otherwise not given): A word
  distinguishing devices belonging to the same person.
"""

def public_server(*, X, Y, public_key, dns_on_vpn=None, public_dns, description=None, prefix_length=None):
    """Used in the config function to define a public server."""
    assert public_dns!=None, "public_dns must be given for a public server."
    host_definition(
        X=X,
        Y=Y,
        public_key=public_key,
        dns_on_vpn=dns_on_vpn,
        public_dns=public_dns,
        description=dns_on_vpn if description==None else description,
        person=None,
        device=None,
        prefix_length=prefix_length,
        enabled=True)

def private_server(*, X, Y, public_key, dns_on_vpn=None, description=None):
    """Used in the config function to define a private server."""
    host_definition(
        X=X,
        Y=Y,
        public_key=public_key,
        dns_on_vpn=dns_on_vpn,
        public_dns=None,
        description=dns_on_vpn if description==None else description,
        person=None,
        device=None,
        prefix_length=None,
        enabled=True)

def personal(*, X, Y, public_key, dns_on_vpn=None, person, device=None):
    """Used in the config function to define a personal device."""
    if Y==0 or Y==1:
        raise ValueError(f"Y={Y} is not suitable for a personal device.")
    if not isinstance(person, str) or not len(person) > 0:
        raise ValueError("person must be a non-empty string.")
    if not isinstance(device, str) or not len(person) > 0:
        raise ValueError("device must be a non-empty string.")
    host_definition(
        X=X,
        Y=Y,
        public_key=public_key,
        dns_on_vpn=dns_on_vpn,
        public_dns=None,
        description=f"{person}, {device}" if device else person,
        person=person,
        device=device,
        prefix_length=None,
        enabled=True)

def disable(*, X=None, Y=None, public_key=None, dns_on_vpn=None, public_dns=None, description=None, person=None, prefix_length=None, device=None):
    """Used in the config function to disable an IP, public key, dns_on_vpn, public_dns, description, person or person/device combination."""
    host_definition(
        X=X,
        Y=Y,
        public_key=public_key,
        dns_on_vpn=dns_on_vpn,
        public_dns=public_dns,
        description=description,
        person=person,
        device=device,
        prefix_length=prefix_length,
        enabled=False)

def assert_unique(string_list, message):
    seen = set()
    for string in string_list:
        if string in seen:
            raise ValueError(message % string)
        else:
            seen.add(string)

def get_host(spec):
    if spec in host_config_map:
        return host_config_map[spec]
    return None

## Functions for generating configuration file content and tarball

def wireguard_config(public_key, private_key):
    """
    Return contents for the wireguard configuration file.
    """
    host = get_host(public_key)
    if host == None:
        print(f"public_key {public_key} not found in host_configs")
        suggest("- If you have recently added a config line for this host and gotten the PR",
                "  merged, it can take a while before the new configuration is published by",
                "  puppet. Try waiting up to 30 minutes and then run again")
        exit_with_suggestions(1)
    assert(host.enabled)
    config_lines = [
        f"# {params.longname} wireguard config for {host.description}",
        "# On Linux/MacOS, this config can only be used with wg-quick.",
        "[Interface]",
        f"PrivateKey = {private_key}",
        f"#PublicKey = {public_key}",
        f"ListenPort = {params.port}" if host.public_dns!=None else None,
        f"Address = {cidr_address(host)}",
    ]
    for other_host in host_configs:
        # Done't include ourselves as a peer
        if other_host.public_key == public_key: continue
        # Exclude disabled hosts
        if not other_host.enabled: continue
        # There is no point including a peer if the connection can't be established
        # in either direction. In this case, the communication will happen via a
        # bounce server instead.
        if host.public_dns==None and other_host.public_dns==None: continue
        # TODO: We want a combined mesh and server/client network, where packets
        #       are sent directly when possible and otherwise fallback to a
        #       bounce server. Unfortunately, native wireguard has no such
        #       fallback, not even skipping peers that are configured but
        #       disconnected. Our short term solution is to limit connections to
        #       those that involve a bounce server routing for a subnet that
        #       includes the IP (or subnet) of the counterparty. This means that
        #       - most traffic will go via a bounce server.
        #       - no traffic will go via more than one bounce server.
        #       - connectivity can be impacted if a bounce server goes down or
        #         the configuration involving bounce servers changes.
        #       - out-of-sync configs will still work for individual hosts.
        #       Long term, we might want to couple wireguard with a dynamic
        #       routing solution, in order to send data directly when possible,
        #       and gracefully handle reconfiguration and servers going down.
        if not (subnet_is_subset(host, other_host) or
                subnet_is_subset(other_host, host)):
            # A peer that is neither up nor down in the IP subnet binary tree
            # should not be connected while we lack dynamic routing
            continue
        # Configure
        config_lines = [
            *config_lines,
            "",
            f"# {other_host.description}",
            "[Peer]",
            f"PublicKey = {other_host.public_key}",
            f"Endpoint = {other_host.public_dns}:{params.port}" if other_host.public_dns!=None else None,
            f"AllowedIPs = {cidr_allowedIPs(other_host)}",
            "PersistentKeepalive = 25",
        ]
    config_lines = [line for line in config_lines if line != None]
    config = "\n".join(config_lines)
    return config

def ip_mask_int(host):
    prefix = host.prefix_length or 32
    IP=(10 * 0x1000000) + (params.subnet * 0x10000)+ (host.X * 0x100) + host.Y
    mask=(0xFFFFFFFF << (32-prefix)) & 0xFFFFFFFF
    return (IP, mask)

def subnet_is_subset(host1, host2):
    (IP1, mask1)=ip_mask_int(host1)
    (IP2, mask2)=ip_mask_int(host2)
    return (mask1 & mask2) == mask2 and (IP1 & mask2) == (IP2 & mask2)

def cidr_address(host):
    return ip(host) + "/16"

def cidr_allowedIPs(host):
    (IP, mask) = ip_mask_int(host)
    effectiveIP=IP & mask
    return f"{effectiveIP >> 24}.{(effectiveIP >> 16) & 0xFF}.{(effectiveIP >> 8) & 0xFF}.{effectiveIP & 0xFF}/{host.prefix_length or 32}"

def sane_filename(x):
    ret=""
    trans_from="ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ ,._=&/*():#;-"
    trans_to=  "AAAAAAACEEEEIIIIDNOOOOOOUUUUYTsaaaaaaaceeeeiiiidnoooooouuuuyty_,._=_____._,-"
    assert(len(trans_from)==len(trans_to))
    for c in x:
        if re.match(r'[A-Za-z0-9_-]', c):
            ret+=c
            continue
        if c in trans_from:
            ret+=trans_to[trans_from.index(c)]
            continue
        raise Exception("Bug")
    ret=ret[:64]
    assert(re.match(r'^[A-Za-z0-9_,.=-]{1,64}$', ret))
    return ret

def make_config_tarball(host):
    tmpdir = tempfile.mkdtemp()
    short_desc=sane_filename(f"{params.longname} configuration for {host.description}")
    dest_count=0
    dirname=""
    dest_filename=""
    while True:
        dirname=f"{short_desc}{'-'+str(dest_count) if dest_count else ''}"
        dest_filename=f"{dirname}.tar.gz"
        if not exists(dest_filename): break
        dest_count+=1
    try:
        if not shBool("which tar"):
            print("tar is not installed. Install it and rerun.")
            exit_with_suggestions(1)
        save_curdir=os.getcwd()
        try:
            print(f"cd {tmpdir}")
            os.chdir(tmpdir)
            sh(f"mkdir {dirname}")
            conffile=f"{dirname}/{params.interface}.conf"
            write(conffile,
                  wireguard_config(host.public_key, '<PRIVATE_KEY>'),
                  is_secret=False)
            print(f"Wrote {conffile}")
            readmefile=f"{dirname}/README.txt"
            readmetext=f"""
                On Linux and MacOS, it is often enough to
                - Copy {params.interface}.conf to /etc/wireguard directory, set permission to 600 and owner to root
                - Edit /etc/wireguard/{params.interface}.conf to insert the private key in the designated place
                - Run wg-quick up {params.interface}

                On Windows, you probably already created a tunnel in order to generate a key
                pair. Edit this tunnel to have the contents of {params.interface}.conf, except
                the private key line which should remain as before.
            """
            write(readmefile, textwrap.dedent(readmetext), is_secret=False)
            print(f"Wrote {readmefile}")
            sh(f"tar czf {dest_filename} {dirname}")
        finally:
            print(f"cd {save_curdir}")
            os.chdir(save_curdir)
        sh(f"cp {tmpdir}/{dest_filename} .")
    finally:
        sh(f"rm -rf {tmpdir}")
    print()
    print(f"Wrote {dest_filename}")

## Functions for configuring this device

def get_system():
    is_linux = platform.system() == "Linux"
    is_macos = platform.system() == "Darwin"
    is_windows = platform.system() == "Windows"
    is_wsl = False
    is_wsl1 = False
    is_wsl2 = False
    if is_linux:
        is_wsl = 'microsoft' in os.uname().release.lower()
        os.uname().release.split('-')[0].split('.')
        kernel_version = os.uname().release.split('-')[0].split('.')
        (kernel_major, kernel_minor) = (int(kernel_version[0]), int(kernel_version[1]))
        is_ge_4_19 = kernel_major > 4 or (kernel_major == 4 and kernel_minor >= 19)
        has_wsl_interop = exists('/run/WSL')
        is_wsl1 = is_wsl and not has_wsl_interop and not is_ge_4_19
        is_wsl2 = has_wsl_interop and is_ge_4_19
        if ((is_wsl1 and is_wsl2) or
            (is_wsl and not is_wsl1 and not is_wsl2) or
            (not is_wsl and (is_wsl1 or is_wsl2))):
            raise Exception(f"Weird system: wsl={is_wsl}, wsl1={is_wsl1}, wsl2={is_wsl2}, release={os.uname().release}, has_wsl_interop={has_wsl_interop}, is_ge_4_19={is_ge_4_19}")
    return (is_linux, is_macos, is_windows, is_wsl, is_wsl1, is_wsl2)
(linux, macos, windows, wsl, wsl1, wsl2) = get_system()

def setup():
    if wsl1:
        print("WSL 1 not supported.")
        suggest("- Convert to WSL 2:",
                "    wsl --list --verbose",
                "    wsl --set-version <Name> 2",
                "    wsl --set-default-version 2",
                "  Also see https://aka.ms/wsl2-install")
        exit_with_suggestions(1)
    if not linux and not macos:
        raise Exception(f"Linux and MacOS supported, but not {platform.system()}.")
    privkeyfile=params.privkeyfile
    pubkeyfile=params.pubkeyfile
    config_file=params.config_file
    interface=params.interface
    # Ensure root
    if os.geteuid()!=0:
        print("This script must be run as root.")
        suggest("- Try the one-line setup:",
                f"    {params.oneline_setup}")
        exit_with_suggestions(1)
    # Ensure wireguard is installed
    ensure_wireguard_installed()
    # Create and check keys
    if not exists(privkeyfile):
        print(f"Need to create {privkeyfile}")
        write(privkeyfile, shOut("wg genkey"), is_secret=True)
        print(f"Created {privkeyfile}")
    if not exists(pubkeyfile):
        print(f"Need to create {pubkeyfile}")
        write(pubkeyfile, shOut(f"wg pubkey < {privkeyfile}"), is_secret=False)
        print(f"Created {pubkeyfile}")
    public_key=read(pubkeyfile)
    expected_public_key=shOut(f"wg pubkey < {privkeyfile}")
    if public_key != expected_public_key:
        print(f"Mismatched keys. The public key in {pubkeyfile} was not generated from the private key in {privkeyfile}.")
        suggest("- Mismatched keys is a weird thing. If you are sure that you want to keep your",
                "  current private key, you can just delete the public key:",
                f"    sudo rm -f {pubkeyfile}",
                "  otherwise you might want to delete both of them:",
                f"    sudo rm -f {pubkeyfile} {privkeyfile}")
        exit_with_suggestions(1)
    print(f"This host is identified by its public key, which is: {public_key}")
    # Get config
    host = get_host(public_key)
    # If config does not exist, help user create it
    if host == None:
        print("This host (identified by public key) is not found in the VPN configuration.")
        print("Before you can join the VPN, you (or someone else) need to create a PR to add")
        print("your host to the VPN configuration.")
        if os.isatty(1):
            if os.isatty(0):
                interactive_setup_help_pr(public_key=public_key)
            else:
                # stdin is not a tty even though stdout is. This indicates
                # piping, which is used in the one-line setup.
                if params.launched_from == "deployed":
                    try_interactive_setup_with_parent_tty(public_key=public_key)
                else:
                    suggest("- To enable interactive questions, try running this script without using shell",
                            "  pipes.")
        else:
            print("Cannot find host config and cannot use terminal for interactive setup help.")
            if params.launched_from == "deployed":
                suggest("- Try the one-line setup:",
                        f"    {params.oneline_setup}")
            else:
                suggest("- To enable interactive questions, try running this script in an interactive")
                suggest("  shell.")
        exit_with_suggestions(1)
    # Show config
    print("VPN configuration for this device:")
    print(host)
    # Teardown if applicable
    if not host.enabled:
        teardown(host)
        return
    # Create or sync config file
    private_key=read(privkeyfile)
    if not isinstance(private_key, str) or not wg_key_pattern.match(private_key):
        raise ValueError(f"Private key read from {privkeyfile} is not a valid WireGuard public key.")
    new_config=wireguard_config(public_key, private_key)
    prev_config=read(config_file)
    if new_config == prev_config:
        print(f"Config file {config_file} is up to date")
    else:
        print(f"Attempting to update {config_file}")
        write(config_file, new_config, is_secret=True)
        print(f"Updated {config_file}")
    # Create or sync interface
    def up():
        print(f"Bringing interface {interface} up.")
        sh(f"{wg_quick_cmd()} up {interface}")
        print("...done")
    was_already_up = is_interface_up()
    if not was_already_up:
        up()
    # On Linux, if the network interface is correctly configured we can
    # synchronize the wireguard configuration without interruption, otherwise we
    # need to bring the interface down and up.
    # On MacOS, wg-quick unfortunately and unnecessarily populates the routing
    # table with one entry per peer, so we always bring it down and up.
    cansync = False
    try:
        if linux and was_already_up and shBool("which ifconfig"):
            interface_config=shOut(f"ifconfig {interface}")
            if not interface_config:
                print("Can't see interface configuration, better bring it down and up.")
            elif f"inet {ip(host)}" in interface_config and (
                    "netmask 255.255.0.0" in interface_config or
                    "netmask 0xffff0000" in interface_config):
                cansync = True
            else:
                print(f"Looks like interface {interface} has wrong IP, so we need to bring it down and up.")
    except Exception as e:
        print(f"Trouble determining whether interface {interface} has correct IP, so we bring it down and up.")
        pass
    need_down_up = was_already_up and not cansync
    if cansync:
        try:
            print(f"Attempting to sync config for interface {interface}")
            assert linux
            sh(f"bash -c 'wg syncconf {interface} <({wg_quick_cmd()} strip {interface})'")
            print("...done")
        except:
            print(f"Syncing failed, so instead we bring interface {interface} down and up.")
            need_down_up=was_already_up
    if need_down_up:
        print(f"Bringing interface {interface} down.")
        sh(f"{wg_quick_cmd()} down {interface}")
        print("...done")
        up()
    # Install system service
    print(f"Installing system service for interface {interface}")
    if linux:
        if init_is_systemd():
            sh(f"systemctl enable wg-quick@{interface}")
            # The wg-quick service waits for the network to become available.
            # However, on Linux this does not guarantee that DNS is available.
            # Our wireguard config files make use of public DNS names for peer
            # endpoints. Also, it appears that when wireguard fails to resolve a
            # DNS name at startup, it reports this as a *parsing* error and
            # fails to bring up the interface. Starting the wg-quick service
            # also is not retried. As a result, if we don't explicitly wait for
            # DNS to become available, a Linux host may fail permanently to join
            # the VPN.
            overridedir="/etc/systemd/system/wg-quick@.service.d"
            override_conf=f"{overridedir}/override.conf"
            # We want to fail if /etc/systemd/system does not exist. Therefore,
            # we do not use `mkdir -p`.
            if not exists(overridedir):
                sh(f"mkdir {overridedir}")
            bash=find_bash_4_or_newer()
            curl=shOut("which curl")
            print(f"Writing {override_conf}")
            write(override_conf,
                  textwrap.dedent(f"""
                    [Service]
                    # Wait for DNS to become available
                    ExecStartPre={bash} -c 'for a in $(seq 300); do if {curl} -so /dev/null http://one.one.one.one; then break; else sleep 1; fi; done'
                  """),
                  is_secret=False)
        else:
            print("WARNING: Could not find a way to install a system service. You will need to")
            print("start and stop the WireGuard interface manually or find a way to automate it.")
            suggest("- If you have a way to run programs at startup, add the following line:",
                    f"    {wg_quick_cmd()} up {params.interface}")
            suggest("- To manually bring down the interface, you can run:",
                    f"    {wg_quick_cmd()} down {params.interface}")
    if macos:
        macos_enable_service()
    print("...done")
    # Open firewall port if we are a public server
    if host.public_dns != None:
        open_port()
        # If the port is left open when no service is listening, a port scan can
        # reveal that something is special about this port. If the ufw default
        # policy is deny (rather than reject), most UDP ports will not give any
        # reply when prodded. An open port with wireguard listening on it will
        # similarly give no reply to unauthenticated traffic, meaning that the
        # wireguard server cannot be discovered by means of port scan. However,
        # if the wireguard application is down but the port open, the kernel
        # will issue an ICMP message that allows a port scanner to detect a
        # difference. Using nmap, UDP ports that are either denied or open with
        # a wireguard server listening will be reported as "open|filtered" while
        # an open port without any service listening will be reported as
        # "closed".
        #
        # TODO: It would be better if the ICMP messages could also be filtered,
        # so that the open port isn't revealed if the server goes down
        # temporarily. Until we fix that, all we can do is warn about it.
        suggest(f"- WARNING: If the wireguard interface {params.interface} goes down, the wireguard",
                f"  port {params.port} could become discoverable by port scan. If you manually bring down",
                "  the interface, be sure to also close the port.")
        suggest("- To make sure that your wireguard port is not discoverable by a port scan, try",
                "  running this from another machine:",
                f"    nmap -sT -sU -p {params.port-2}-{params.port+2} {host.public_dns}",
                f"  Make sure that the output for port {params.port} looks the same as for the",
                "  other ports.")
    # Make sure routing is turned on if we are a bounce server
    if host.prefix_length != None and host.prefix_length < 32:
        forwarding=read('/proc/sys/net/ipv4/ip_forward')
        if forwarding=="1":
            print("IPv4 forwarding is already turned on.")
        else:
            print("Trying to configure as bounce server, but IPv4 forwarding turned off.")
            suggest("- To temporarily turn on IPv4 forwarding:",
                    "    echo 1 > /proc/sys/net/ipv4/ip_forward")
            suggest("- To turn on IPv4 forwarding on startup:",
                    "    echo 'net.ipv4.ip_forward = 1' >> /etc/sysctl.conf")
            exit_with_suggestions(1)
        open_routing(host)
    if not 'RUNNING_UNDER_PUPPET' in os.environ:
        suggest("- Setup is complete, but over time your configuration might get out of sync. For",
                "  the most part, the VPN will still work. You might want to synchronize the",
                "  config occasionally, or automatically, or when you experience any trouble.",
                "  Make note of this one-liner:",
                f"    {params.oneline_setup}")

def ensure_wireguard_installed():
    if not shBool("which wg"):
        print("Wireguard is not installed. Install it and rerun.")
        if shBool("which yum"):
            suggest("- Try installing wireguard this way:",
                    "    sudo yum install -y epel-release elrepo-release",
                    "    sudo yum install -y kmod-wireguard wireguard-tools")
        if shBool("which apt-get"):
            suggest("- Try installing wireguard this way:",
                    "    sudo apt-get -y update",
                    "    sudo apt-get -y install wireguard")
        if shBool("which brew"):
            suggest("- Try installing wireguard this way:",
                    "    brew update",
                    "    brew install wireguard-tools")
        exit_with_suggestions(1)
    if linux and not wsl1 and not wsl2 and not "wireguard" in shOut("lsmod"):
        print("Wireguard seems to be installed correctly but the wireguard kernel module is not")
        print("loaded. Install/load the kernel module and rerun.")
        if shBool("which modprobe"):
            suggest("- If the wireguard kernel module is installed but not loaded, try:",
                    "    sudo modprobe wireguard",
                    "  If that does not work it could be due to a kernel update; try rebooting.")
        exit_with_suggestions(1)

def read(file_path):
    if not exists(file_path): return ""
    with open(file_path, 'r', encoding='utf-8') as file:
        return file.read().strip()

def write(file_path, content, *, is_secret=True):
    permissions = 0o600 if is_secret else 0o644
    if not content.endswith('\n'):
        content += '\n'
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    try:
        fd = os.open(file_path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, permissions)
        os.fchmod(fd, permissions)
        with os.fdopen(fd, 'w', encoding='utf-8') as file:
            file.write(content)
    except:
        os.close(fd)

def sh(cmd):
    """Execute command without capturing output. Raise an exception if it fails."""
    if '|' in cmd: raise ValueError("Piping is forbidden")
    print(cmd)
    subprocess.run(cmd, shell=True, check=True)

def shOut(cmd):
    """Return stdout if command was successful, otherwise raise an Exception."""
    if '|' in cmd: raise ValueError("Piping is forbidden")
    try:
        result = subprocess.run(
            # To be compatible with Python 3.6, we need to avoid arguments `text`,
            # `capture_output` and `encoding`.
            cmd,
            shell=(isinstance(cmd,str)),
            check=True,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        # Accessing stdout and stderr from the exception object
        print("COMMAND:")
        print(cmd)
        print("STDOUT:")
        print(e.stdout)
        print("STDERR:")
        print(e.stderr)
        raise e
    return result.stdout.strip()

def shBool(cmd):
    """Return True for exit code 0, False for 1, otherwise raise an Exception."""
    if '|' in cmd: raise ValueError("Piping is forbidden")
    null=subprocess.DEVNULL
    result = subprocess.run(cmd, shell=True, check=False, stdout=null, stderr=null)
    if result.returncode==0: return True
    if result.returncode==1: return False
    raise Exception(f"Expected exit code 0 or 1, but got {result.returncode}")

def open_port():
    p=params.port
    if shBool('which ufw'):
        print(f"Opening UDP port {p} with ufw.")
        sh(f"ufw allow {p}/udp comment '{params.longname}'")
        print("...done")
        return
    print()
    raise Exception(f"ufw not installed, don't know how to open port {p}")

def close_port():
    p=params.port
    if not shBool('which ufw'):
        # If ufw is not installed, then we don't need to remove any rule.
        return
    print(f"Closing UDP port {p} with ufw.")
    sh(f"ufw delete allow {p}/udp")
    print("...done")

def open_routing(host):
    if shBool('which ufw'):
        route_net=cidr_allowedIPs(host)
        print(f"Allowing routing for {params.code}...")
        sh(f"ufw route allow from {route_net} to {route_net} comment 'Bounce server routing for {params.longname}'")
        print("...done")
        return
    print()
    raise Exception(f"ufw not installed, don't know how to allow routing for {params.code}")

def close_routing(host):
    if not shBool('which ufw'):
        # If ufw is not installed, then we don't need to remove any rule.
        return
    route_net=cidr_allowedIPs(host)
    print(f"Allowing routing for {params.code}...")
    sh(f"ufw route delete allow from {route_net} to {route_net}")
    print("...done")

def teardown(host):
    interface=params.interface
    config_file=params.config_file
    if is_interface_up():
        print("Bringing interface {interface} down...")
        sh(f"{wg_quick_cmd()} down {interface}")
        print("...done")
    if exists(config_file):
        print(f"Removing config file {config_file}")
        sh(f"rm -r {config_file}")
        print("...done")
    print(f"Uninstalling system service for interface {interface}")
    if linux:
        if init_is_systemd():
            sh(f"systemctl disable wg-quick@{interface}")
    if macos:
        macos_disable_service()
    if exists(params.setupstatefile):
        sh(f"rm -f {params.setupstatefile}")
    print("...done")
    close_port()
    close_routing(host)
    print(f"Teardown complete, this host is no longer connected to {params.longname}.")
    if exists(params.pubkeyfile):
        suggest("- The public key was not deleted. If you are sure that you want it gone, you can",
                "run:",
                f"    rm -f {params.pubkeyfile}")
    if exists(params.privkeyfile):
        suggest("- The private key was not deleted and it is recommended that you keep it. If you",
                "  are sure that you want it gone, you can run:",
                f"    rm -f {params.privkeyfile}")

def is_interface_up():
    interface=params.interface
    if linux:
        # Also works on WSL 2
        return shBool(f"ip link show {interface}")
    if macos:
        name_file=f"/var/run/wireguard/{interface}.name"
        if not exists(name_file): return False
        real_interface=read(name_file)
        return shBool(f"ifconfig {real_interface}")

def init_is_systemd():
    return shOut("ps -p 1").endswith("systemd")

def macos_enable_service():
    # wg-quick requires bash 4+
    bash=find_bash_4_or_newer()
    curl=shOut("which curl")
    service_file_content=f"""
    <?xml version="1.0" encoding="UTF-8"?>
    <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
    <plist version="1.0">
    <dict>
        <key>Label</key>
        <string>com.user.wg-quick.{params.interface}</string>
        <key>ProgramArguments</key>
        <array>
            <string>{bash}</string>
            <string>-c</string>
            <string>until {curl} -fIs {params.livenessuri} >/dev/null; do sleep 5; done; {wg_quick_cmd()} up {params.interface}</string>
        </array>
        <key>RunAtLoad</key>
        <true/>
    </dict>
    </plist>
    """
    service_name=f"com.user.wg-quick.{params.interface}"
    service_file=f"/Library/LaunchDaemons/{service_name}.plist"
    write(service_file, service_file_content, is_secret=False)
    sh(f"launchctl enable system/{service_name}")

@lru_cache()
def wg_quick_cmd():
    return f"{find_bash_4_or_newer()} {shOut('which wg-quick')}"

def find_bash_4_or_newer():
    for path in shOut("which -a bash").split('\n'):
        version_text=shOut(f"{path} --version")
        if re.match(r".* version [456789]\..*", version_text):
            return path
    print("You need to install bash version 4 or later.")
    if shBool("which brew"):
        suggest("- Try installing bash this way:",
                "    brew update",
                "    brew install bash")
    exit_with_suggestions(1)

def macos_disable_service():
    service_name=f"com.user.wg-quick.{params.interface}"
    service_file=f"/Library/LaunchDaemons/{service_name}.plist"
    sh(f"launchctl disable system/{service_name}")
    sh(f"rm -f {service_file}")

def interactive_setup_help_pr(*, public_key, is_recursive=False, on_this_machine=True):
    global segments, host_configs
    def ask(question, validate=None):
        while True:
            user_input = input(question + " ")
            print()
            if validate is None or validate(user_input):
                return user_input
            else:
                print("Invalid input, please try again.")
    def askYN(question):
        answer = ask(question + " [y/n]", lambda x: x.strip().lower() in ["yes", "y", "no", "n"])
        return answer.strip().lower() in ["yes", "y"]
    def print_cut_line(line):
        cut="== 8< =============="
        cutline=(cut * (len(line)//len(cut)) + ("="*len(cut)))[:len(line)]
        print(cutline)
        print(line)
        print(cutline)
    if on_this_machine:
        previous_config_line=read_state_var("suggested_config_line")
        if previous_config_line and not is_recursive:
            print("It looks like you've already produced a config line for your host:")
            print_cut_line(previous_config_line)
            if(askYN("Does this look correct?")):
                print("We're still waiting on this to get merged and published. Rerun later.")
                suggest("- If you have recently added a config line for this host and gotten the PR",
                        "  merged, it can take a while before the new configuration is published by",
                        "  puppet. Try waiting up to 30 minutes and then run again.")
                return
            else:
                print("Then let's correct it.")
                print()
        print("You'll now get a few interactive questions to help you produce a config line for this machine.")
        print()
    else:
        print("You'll now get a few interactive questions to help you produce a config line for a new host.")
        print()
        public_key=ask("What is the public key of the new host?", lambda x: wg_key_pattern.match(x))
        existing_host=get_host(public_key)
        if existing_host and existing_host.public_key==public_key:
            print("This public key is already in use by a host in the config.")
            exit_with_suggestions(1)
    is_personal = askYN("Is this a personal device of an internal employee (such as a single-user laptop or development VM)?")
    occupied_XY = []
    for host in host_configs:
        if host.X == None and host.Y == None: continue
        occupied_XY.append((host.X, host.Y))
    X=None
    Y=None
    if is_personal:
        for segment in segments:
            if not segment.available:
                continue
            if not segment.personal:
                continue
            X=segment.X
            for maybe_Y in range(2, 255): # meaning from 2 to 254
                if (X, maybe_Y) not in occupied_XY:
                    Y=maybe_Y
                    break
            if Y is not None: break
    else:
        segment_candidates=[
            segment for segment in segments
            if segment.available and not segment.personal]
        print("Available network segments are:")
        for segment in segment_candidates:
            print(f"{segment.X}: {segment.description}")
        X=int(ask("Which network segment do you want to join? (X)",
                  lambda x: x.isdigit() and int(x) in [segment.X for segment in segment_candidates]))
        for maybe_Y in range(2, 255): # meaning from 2 to 254
            if (X, maybe_Y) not in occupied_XY:
                Y=maybe_Y
                break
        if Y==None:
            raise Exception(f"No available address in segment {X}")
    fun=personal if is_personal else (
        public_server if askYN("Is this host directly routable from the public internet (i.e. has its own public IP)?") else private_server)
    args=[("X", X), ("Y", Y)]
    args.append(("public_key", public_key))
    public_dns = (ask("Enter the public DNS name for this host:",
                      lambda x: dns_name_pattern.match(x))
                  if fun==public_server else None)
    dns_on_vpn=None
    if askYN("Do you want to create a DNS name for reaching this host over the VPN?" +
             (" (usually not for personal devices)" if fun==personal else '')):
        print(f"The DNS name for this host will be <LABEL>.{params.zone}")
        dns_on_vpn=ask("Enter DNS label:",
                       lambda x: dns_label_pattern.match(x))
    if dns_on_vpn is not None: args.append(("dns_on_vpn", dns_on_vpn))
    if public_dns is not None: args.append(("public_dns", public_dns))
    if fun!=personal and (
            dns_on_vpn is None or
            askYN("A description is not necessary for this host. Do you still want to provide one?")):
        args.append(("description",
                       ask("Enter description for this host:",
                           lambda x: description_pattern.match(x))))
    if fun==personal:
        args.append(("person",
                     ask("Enter your full name:",
                         lambda x: person_pattern.match(x))))
        args.append(("device",
                     ask("Enter a name for this device using only a-z 0-9 (e.g. laptop):",
                         lambda x: device_pattern.match(x))))
    try:
        def addthis():
            global current_segment
            current_segment=X
            fun(**dict(args))
            current_segment=None
        load_and_check_config(extra=addthis)
    except ValueError as e:
        print(f"There's a problem with your configuration: {e.args[0]}")
        print(f"Try again!")
        # Reset the config to have nothing extra and try again.
        load_and_check_config()
        return interactive_setup_help_pr(public_key=public_key, is_recursive=True)
    print("\n\n\n")
    print(f"That looks good! The suggested configuration line for {'this host' if on_this_machine else 'the host in question'} is as follows:")
    print()
    formatted_args=", ".join([f"{name}={repr(val)}" for (name, val) in args])
    line=f"{fun.__name__}({formatted_args})"
    if on_this_machine:
        write_state_var("suggested_config_line", line)
    print_cut_line(line)
    print()
    print("You will need to get this line into the puppet repo at")
    print(f"{params.repo_uri} into file")
    print(params.repo_dir_for_config_script + params.config_script_filename)
    if on_this_machine:
        print("After that, you can rerun this script to continue setup.")
    else:
        print("After that, you can produce the configuration by running")
        print(f"  {sys.argv[0]} --config {public_key}")

def read_state_var(var):
    return json.loads(read(params.setupstatefile) or "{}").get(var)

def write_state_var(var, val):
    write(params.setupstatefile,
          json.dumps({**json.loads(read(params.setupstatefile) or "{}"),
                      var: val}))

def try_interactive_setup_with_parent_tty(*, public_key):
    for parent_stdin_path in stdin_for_process_hierarchy(os.getpid()):
        if not parent_stdin_path or not (
                parent_stdin_path.startswith("/dev/tty") or
                parent_stdin_path.startswith("/dev/pts/")):
            continue
        with open(parent_stdin_path, 'r') as effective_stdin:
            if not os.isatty(effective_stdin.fileno()):
                continue
            orig_stdin = sys.stdin
            try:
                sys.stdin = effective_stdin
                return interactive_setup_help_pr(public_key=public_key)
            finally:
                sys.stdin = orig_stdin
    print("Detected piped input but could not find a terminal, which is necessary for")
    print("interactive questions.")
    suggest("- Try the one-line setup:",
            f"    {params.oneline_setup}")

def stdin_for_process_hierarchy(pid):
    if not isinstance(pid, int) or pid <= 1:
        return []
    # Find stdin
    stdin_path=None
    try:
        open_files=shOut(f"lsof -p {pid} -Ffn")
    except subprocess.CalledProcessError as e:
        print(e)
        return []
    lines = open_files.splitlines()
    for i, line in enumerate(lines):
        if line == "f0" and i + 1 < len(lines) and lines[i + 1].startswith("n"):
            stdin_path=lines[i + 1][1:]
            break
    ppid=int(shOut(f"ps -o ppid= -p {pid}") or "0")
    if stdin_path and exists(stdin_path):
        stdin_path=shOut(["readlink", "-f", stdin_path])
    else:
        stdin_path=None
    if stdin_path:
        return [stdin_path, *stdin_for_process_hierarchy(ppid)]
    return stdin_for_process_hierarchy(ppid)

## Functions for updating DNS zones

def update_dns():
    change_requests=[]
    expected = expected_dns_records()
    actual = get_dns_records()
    def tomap(records):
        ret={}
        for record in records:
            key="{Name} {Type} {TTL}".format(**record)
            if key in ret:
                raise Exception(f"Duplicate key '{key}' when investigating DNS records.")
            ret[key]=record
        return ret
    expected_map=tomap(expected)
    actual_map=tomap(actual)
    changes=[]
    print()
    print(f"Changes to be applied to {params.zone}:")
    for key in actual_map:
        if not key in expected_map:
            record=actual_map[key]
            print(f"- DELETE {record}")
            changes.append({
                "Action": "DELETE",
                "ResourceRecordSet": record
            })
    for key in expected_map:
        doupsert=False
        if key in actual_map:
            expected_rr=expected_map[key]['ResourceRecords']
            actual_rr=actual_map[key]['ResourceRecords']
            def sort_key(d): return tuple(sorted(d.items()))
            if sorted(expected_rr, key=sort_key)!=sorted(actual_rr, key=sort_key):
                doupsert=True
        else:
            doupsert=True
        if doupsert:
            print(f"- UPSERT {expected_map[key]}")
            changes.append({
                "Action": "UPSERT",
                "ResourceRecordSet": expected_map[key]
            })
    if not changes:
        print(f"None, zone {params.zone} is already up to date.")
        return
    while changes:
        this_batch=changes[:100]
        changes=changes[100:]
        print()
        print(f"Requesting {len(this_batch)} change{pl(len(this_batch),'','s')} in zone {params.zone} ...")
        request_body={"Changes": this_batch}
        change_request = shOut(f"{route53_cmd()} change-resource-record-sets --hosted-zone-id {params.aws_route53_zone_id} --change-batch '{json.dumps(request_body)}'")
        print("...done")
        change_id = json.loads(change_request)['ChangeInfo']['Id']
        change_requests.append((change_id, len(this_batch)))
    while change_requests:
        count=0
        for (change_id, nof_changes) in change_requests:
            count+=nof_changes
        print()
        print(f"{count} change{pl(count,'','s')} remain{pl(count,'s','')} to propagate in zone {params.zone} ...")
        time.sleep(5)
        remaining_change_requests=[]
        for (change_id, nof_changes) in change_requests:
            response=shOut(f"{route53_cmd()} get-change --id {change_id}")
            if json.loads(response)['ChangeInfo']['Status']=='PENDING':
                remaining_change_requests.append((change_id, nof_changes))
        change_requests=remaining_change_requests
    print()
    print("All changes propagated.")

@lru_cache()
def route53_cmd():
    akid=read_state_var("AWS_ACCESS_KEY_ID")
    sak=read_state_var("AWS_SECRET_ACCESS_KEY")
    return ((f"AWS_ACCESS_KEY_ID={akid} " if akid else "") +
            (f"AWS_SECRET_ACCESS_KEY={sak} " if sak else "") +
            "aws route53")

def pl(count, singular, plural): return singular if count==1 else plural

def expected_dns_records():
    records=[]
    for host in host_configs:
        if host.enabled and host.dns_on_vpn:
            records.append({
                "Name": f"{host.dns_on_vpn}.{params.zone}.",
                "Type": "A",
                "TTL": 300,
                "ResourceRecords": [
                    {
                        "Value": ip(host)
                    }
                ]
            })
    return records

def get_dns_records():
    print()
    print(f"Getting DNS records for zone {params.zone} ...")
    response = json.loads(shOut(f"{route53_cmd()} list-resource-record-sets --hosted-zone-id {params.aws_route53_zone_id}"))
    records = [ rec for rec in response['ResourceRecordSets'] if rec['Type'] not in ["NS", "SOA"]]
    print(f"...done, got {len(records)} records.")
    return records

# Functions for rendering config

def render_config(*, render_format):
    host_records = [
        {
            "dns": f"{host.dns_on_vpn}.{params.zone}" if host.dns_on_vpn else "",
            "ip": ip(host),
            "description": host.description,
        }
        for host in host_configs
        if host.enabled
    ]
    if render_format=="json":
        print(json.dumps({"hosts": host_records}, indent=2))
        return
    if render_format=="html":
        recs=('\n                  ').join([
            f"<tr><td>{r['dns']}</td><td>{r['ip']}</td><td>{html.escape(r['description'])}</td></tr>"
            for r in host_records
        ])
        print(textwrap.dedent(f"""
          <!DOCTYPE html>
          <html lang="en">
            <head>
              <meta charset="utf-8">
              <title>Host records for {params.longname}</title>
              <link rel="stylesheet"
                    crossorigin="anonymous"
                    href="https://cdn.jsdelivr.net/npm/water.css@2/out/water.css">
            </head>
            <body>
              <h1>Host records for {params.longname}</h1>
              <table style="border-collapse: collapse; border: 1px solid black;">
                <thead>
                  <tr><th>DNS name</th><th>IP address</th><th>Description</th></tr>
                </thead>
                <tbody>
                  {recs}
                </tbody>
              </table>
            </body>
          </html>
        """).strip())
        return
    raise Exception(f"Unknown render format: {render_format}")

## Functions for suggestions at exit

suggestions=[]
def suggest(*lines):
    suggestions.append("\n".join(lines))
def print_suggestions():
    if suggestions:
        print()
        print("Suggestions:")
        for suggestion in suggestions:
            print(suggestion)
def exit_with_suggestions(ec):
    print_suggestions()
    sys.exit(ec)
#!/usr/bin/env python3
# -*- coding: utf-8 -*-

FILETOP_DOCUMENTATION="""

=== HOW TO JOIN THE VPN ===

Hopsworks devnet VPN can be used by anyone internally.
Run this one-line setup on the machine to be joined:
bash -c 'set -o pipefail; curl -fLsS https://repo.hops.works/vpn/oCetFIEGcL/devnet.py | sudo python3'

The first time you run the setup, you will be given the config line that needs
to be added in this file, and instructions for how to continue.

=== HOW TO SYNCHRONIZE THE VPN CONFIGURATION ON YOUR MACHINE ===

Just run the one-line setup again.
Usually the VPN works fine even if your configuration is out of sync.

=== HOW TO DISABLE A HOST ===

When disabling a host, change the function name (public_server, private_server
or personal) to `disable` and keep the same arguments. The `disable` function
can also be used to disable a single IP, public key, dns_on_vpn, public_dns,
description, person or person/device combination. Examples:
- disable(X=7, Y=8)
- disable(public_key="AaAaAaAaAaAaAaAaAaAaAaAaAaAaAaAaAaAaAaAaAaA=")
- disable(dns_on_vpn="oldhost")
- disable(public_dns="oldhost.hops.works")
- disable(description="The old host")
- disable(person="Joe Former") # This makes sure no hosts are configured for
                               # this person, no matter whether device is
                               # specified.
- disable(person="Joe Former", device="laptop")
"""



### BEGINNING OF CONFIG

def config():

    ### Segments 0-7 Named resources, cloud/baremetal

    with segment(X=1,  available=True, description="Dev servers (bare-metal devX.hops.works)"):
        public_server(X=1, Y=0, public_key='mFYOIanwcv1I9P2tITA+DQPsdfbkYs3zn7RVoox/AAI=', dns_on_vpn='dev0', public_dns='dev0.hops.works')
        public_server(X=1, Y=1, public_key='I8d+MOfvPCZ07TK5YuwaQW4L4/dvZ//xm2ugvSrHdQ8=', dns_on_vpn='dev1', public_dns='dev1.hops.works')
        public_server(X=1, Y=2, public_key='z+l4YefxhcEVyXgUGp2asQXYFRx6OA+qOYkAHmqaSzw=', dns_on_vpn='dev2', public_dns='dev2.hops.works')
        public_server(X=1, Y=3, public_key='3BoW/fr7pPblB+/cc+/GTA4+QNDSE1mz84GKsoqKNQ0=', dns_on_vpn='dev3', public_dns='dev3.hops.works')
        public_server(X=1, Y=4, public_key='s5ozQhCpdsIBsktOakSyU4O7Jw9ZcR+eqD86thDMjEE=', dns_on_vpn='dev4', public_dns='dev4.hops.works', prefix_length=16, description="Wireguard bounce server")
        public_server(X=1, Y=5, public_key='Ncp5QNAB4pgSLTeKfviP/gCtsJvGHLNir86ZEl3Bk1w=', dns_on_vpn='dev5', public_dns='dev5.hops.works')
        public_server(X=1, Y=6, public_key="cmM7Bgsuh35vr2UuLJ/2DyMqcTi0g8zPEoIEpiMosHA=", dns_on_vpn="dev6", public_dns="dev6.hops.works")

    with segment(X=2, available=True, description="Other infrastructure"):
        public_server(X=2, Y=2, public_key='EsBJegPuB9+zNHeoGeqHf6Jy8/YEYl2lnbPWM2sTSW0=', dns_on_vpn='jenkins', public_dns='jenkins.hops.works', description='jenkins.hops.works and repo.hops.works')
        private_server(X=2, Y=3, public_key='SU7LsTjOWH3ppIf1Me8Lk4OgLx+MsN/mbYhdr8XWTVk=', dns_on_vpn='inferyxvm', description='Inferyx VM at dev0')
        public_server(X=2, Y=4, public_key='Pmr3R9i4vN3hK4X6QQvuY/rl/oy+Mu+m/w2SJO9x40k=', dns_on_vpn='executor0', public_dns='executor0.hops.works', description='jenkins slave executor0.hops.works')

    ### Segments 16-23 Stockholm Office

    with segment(X=16, available=True, description="Stockholm Office, Designated devices (bare-metal, permanent/non-roaming)"):
        disable(X=16, Y=2, public_key='dC0RXcEDlY9jnE/IQHfNPXg2u6pCw3RL0SY3BRF2fCc=', description='thesofa, temporary key')
        disable(X=16, Y=3, public_key='7+mF8lawEeEdv8NCpFAyKqhWdt4U6lY3mjLc3f3yPCI=', description='thesofa, backup connection')
        private_server(X=16, Y=4, public_key='sofa/v22zOIFLtvGgNkD8mhesguM75VW2MMTpndDsCg=', dns_on_vpn='thesofa', description='RonDB team inhouse compute node (thesofa)')
        private_server(X=16, Y=5, public_key='g1PsGVyovJpekU1onj9zo8g0nKtJFjY4o5v+wnSAKmc=', dns_on_vpn='snurran', description='Lab machine with powerful GPU to experiment with Large-Language Models and Hopsworks (snurran)')

    with segment(X=20, available=False, description="Stockholm Office, DHCP (unidentified/roaming devices on WiFi etc.)"): pass

    with segment(X=21, available=False, description="Stockholm Office, DHCP (unidentified/roaming devices on WiFi etc.)"): pass

    ### Segments 72-79

    with segment(X=72, available=False, description="Do not use (to avoid confusion with prodnet subnet)"): pass

    ### Segments 80-87 Personal devices

    with segment(X=80, available=True, description="Personal devices for internal employees (single-user, for example laptops or development VMs)", personal=True):
        personal(X=80, Y=2, public_key='6fiQI3BuFtRUWw+vC7CrV3L5HXJydUvJGUHLSSq+N1M=', person='Axel Svensson', device='lcd')
        personal(X=80, Y=3, public_key="yoHk4Gwt/8UcyA5ignANxA+K35PdK4SG5q/YUtRvkTs=", person="Ray Cunningham", device="macos")
        disable(X=80, Y=4, public_key="aSfDV9EVejjHNm6wlCFZpcepgVI7EBnz1OyWw01cICU=", person="Ray Cunningham", device="manjaro-test")
        disable(X=80, Y=5, public_key='caV5xBoyCK/0zy4QUU/iafxIF6WLxwTkIhu9mGaJrjU=', person='Victor Jouffrey', device="laptop")
        disable(X=80, Y=6, public_key='u8YMGSUWQsJR/6o0f52aZXObH8Z47/mlYP/XAYpwSns=', person='Axel Svensson', device='aws-macos-testing')
        personal(X=80, Y=7, public_key='rkszSyyWiBWuUbSajvndK+r9BIE7VqJeGqDDoItV+DU=', person='Axel Svensson', device='laptop-wsl')
        personal(X=80, Y=8, public_key='2etrRVRz3pCgEq9kFisNf5jTtBHK9S/LUZ3lLoOpOGs=', person='Ray Cunningham', device='littlefastgasson')
        personal(X=80, Y=9, public_key='LzpO1LPJPi7LTdoLPb+RGqiFLQ0e6B6Ik1ZKwsIQEBI=', person='Mikael Ronström', device='macmini')
        personal(X=80, Y=10, public_key='wWyie8+xPMgziJ577uAVn6ydTPUvKaQIMj89yYZdFB4=', person='Antonios Kouzoupis', device='feanorlaptop')
        personal(X=80, Y=11, public_key='GHllOO+qXPVOD9DribcviNX5kojbV7GCcVnmilFZBEc=', dns_on_vpn='fabio-tiramisu', person='Fabio Buso', device='tiramisu')
        personal(X=80, Y=12, public_key='ktsnTuFrhSotbzApit+7STwGYHkbqElTWHDXTK/RsnI=', dns_on_vpn='axel-devvm1', person='Axel Svensson', device='devvm1')
        personal(X=80, Y=13, public_key='X50mNzstGWiKXDrd3qWAsdIZBFHEzDg/zGkk/J8tIGU=', person='Zhao Song', device='zhao')
        personal(X=80, Y=14, public_key='bpaJg4tA3tJ2uxJ6Fdi5c3EhqWxee5pe+03gtJDBT18=', person='Victor Jouffrey', device='mymac')
        personal(X=80, Y=15, public_key='vmcT/kV3BZ3OmLYI9UGVZ/VDtaGFhAj5AkWiG+FxySc=', person='Siyao Liu', device='ysl641')

    with segment(X=87, available=False, description="Do not use (to avoid confusion with devnet subnet)"): pass

    ### Segments 96-103 External parties

    with segment(X=96, available=True, description="External parties, people outside Hopsworks"):
        private_server(X=96, Y=2, public_key='jvW/z9dq4R2rfkzkNh0u8bq4QDrq7p6RBfRrSZdTACQ=', description='Inferyx, Gopal Kumar')
        private_server(X=96, Y=3, public_key='jzV3j6PHbMSiCxGnl5YP3P6JltlP+AivRqU4eCl12H8=', description='Inferyx, Dipak Dose')
        private_server(X=96, Y=4, public_key='E99fPyOSgzai+/sPpDvvwvHAUoebHA7EmBXU3jV32x4=', description='Inferyx, Vaibhav Katkar')
        private_server(X=96, Y=5, public_key='fEugl9n0eI8P72VviB+DueGSCTF9ocoTOKykf9mvEUc=', description='Inferyx, Yogesh Palrecha')
        # 2OP... is a private key exposed via email from Gopal Kumar to Axel
        # Svensson amongst others on 2024-03-11. M8N... is the corresponding
        # public key. Since private and public keys have the same format, we
        # might as well disable both.
        disable(public_key="2OPEhtrubM76B6G77gQYoMGOadgZGnyeRMUTKgWzeGM=")
        disable(public_key="M8NUjeZIexg5Ct1rJXRYxjNNQZX/jCp1v0roMQahExA=")
        private_server(X=96, Y=6, public_key='qcKPPKb1i8OxlBWrHqJdqPGGCH5xEBrKq9xSHimp51Y=', description='Inferyx, Jitender Kumar')


### PARAMETERS

def get_parameters():
    code="devnet"
    # This obfuscation value must match that in ./networks_and_obfuscation.json
    obfuscation="oCetFIEGcL"
    publish_uri=f"https://repo.hops.works/vpn/{obfuscation}"
    this_script_filename="devnet.py"
    return {
        # Used to update DNS entries for the VPN
        "aws_route53_zone_id": "Z06715082P90AC53NIGBC",
        # Short name for the VPN network, 1-10 characters a-z
        "code": code,
        # Configuration, a function calling segment, public_server, private_server, personal and disable.
        "config": config,
        "config_script_filename": this_script_filename,
        "issues_uri": "https://github.com/logicalclocks/puppet/issues",
        "launched_from": "deployed" if 'main' in globals() else "repo",
        # A URI used to check for internet connectivity
        "livenessuri": "http://1.1.1.1",
        # A more descriptive name for the VPN network
        "longname": f"Hopsworks {code} VPN",
        "oneline_setup": f"bash -c 'set -o pipefail; curl -fLsS {publish_uri}/{this_script_filename} | sudo python3'",
        # UDP port that wireguard listens to on public servers
        "port": 17486,
        "repo_dir_for_config_script": "code/modules/vpn/files/",
        "repo_uri": "https://github.com/logicalclocks/puppet",
        # We use the IPv4 block 10.87.0.0/16
        "subnet": 87,
        "zone": f"{code}.hops.works",
    }
parameters=get_parameters()

### END OF CONFIG



assert parameters['oneline_setup'] in FILETOP_DOCUMENTATION
# vpn.py and this file are concatenated before publishing. In that case we
# shouldn't import.
if parameters['launched_from'] == "repo":
    from vpn import main, segment, public_server, private_server, personal, disable
main(**parameters)
